from .base import MriStudy
from nipype.interfaces.fsl import TOPUP, ApplyTOPUP
from nipype.interfaces import fsl
from nipype.interfaces.utility import Merge as merge_lists
from nipype.interfaces.fsl.utils import Merge as fsl_merge
from nipype.interfaces.fsl.epi import PrepareFieldmap
from nipype.interfaces.fsl.preprocess import BET, FUGUE
from arcana.study import ParamSpec, SwitchSpec
from arcana.data import InputFilesetSpec, FilesetSpec, FieldSpec
from arcana.study.base import StudyMetaClass
from banana.citation import fsl_cite
from banana.requirement import fsl_req
from banana.interfaces.custom.motion_correction import (
    PrepareDWI, GenTopupConfigFiles)
from banana.file_format import (
    nifti_gz_format, text_matrix_format,
    par_format, motion_mats_format, dicom_format)
from banana.interfaces.custom.bold import FieldMapTimeInfo
from banana.interfaces.custom.motion_correction import (
    MergeListMotionMat, MotionMatCalculation)

from banana.file_format import STD_IMAGE_FORMATS


class EpiStudy(MriStudy, metaclass=StudyMetaClass):

    add_data_specs = [
        InputFilesetSpec('coreg_ref_wmseg', STD_IMAGE_FORMATS,
                         optional=True),
        InputFilesetSpec('reverse_phase', STD_IMAGE_FORMATS, optional=True),
        InputFilesetSpec('field_map_mag', STD_IMAGE_FORMATS,
                            optional=True),
        InputFilesetSpec('field_map_phase', STD_IMAGE_FORMATS,
                            optional=True),
        FilesetSpec('moco', nifti_gz_format,
                    'intrascan_alignment_pipeline'),
        FilesetSpec('align_mats', motion_mats_format,
                    'intrascan_alignment_pipeline'),
        FilesetSpec('moco_par', par_format,
                    'intrascan_alignment_pipeline'),
        FieldSpec('field_map_delta_te', float,
                  'field_map_time_info_pipeline')]

    add_param_specs = [
        SwitchSpec('bet_robust', True),
        SwitchSpec('linear_coreg_method', 'epireg',
                   (MriStudy.parameter_spec('linear_coreg_method').choices +
                    ('epireg',))),
        ParamSpec('bet_f_threshold', 0.2),
        ParamSpec('bet_reduce_bias', False),
        ParamSpec('fugue_echo_spacing', 0.000275)]

    def linear_coreg_pipeline(self, **name_maps):
        if self.branch('linear_coreg_method', 'epireg'):
            return self._epireg_linear_coreg_pipeline(**name_maps)
        else:
            return super(EpiStudy, self).linear_brain_coreg_pipeline(
                **name_maps)

    def _epireg_linear_coreg_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='linear_coreg',
            desc=("Intra-subjects epi registration improved using white "
                  "matter boundaries."),
            citations=[fsl_cite],
            name_maps=name_maps)

        pipeline.add(
            'epireg',
            fsl.epi.EpiReg(
                out_base='epireg2ref'),
            inputs={
                'epi': ('brain', nifti_gz_format),
                't1_brain': ('coreg_ref_brain', nifti_gz_format),
                't1_head': ('coreg_ref', nifti_gz_format),
                'wmseg': ('coreg_ref_wmseg', nifti_gz_format)},
            outputs={
                'coreg_brain': ('out_file', nifti_gz_format),
                'coreg_matrix': ('epi2str_mat', text_matrix_format)},
            requirements=[fsl_req.v('5.0.9')])

        return pipeline

    def intrascan_alignment_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='MCFLIRT_pipeline',
            desc=("Intra-epi volumes alignment."),
            citations=[fsl_cite],
            name_maps=name_maps)

        mcflirt = pipeline.add(
            'mcflirt',
            fsl.MCFLIRT(
                ref_vol=0,
                save_mats=True,
                save_plots=True,
                output_type='NIFTI_GZ',
                out_file='moco.nii.gz'),
            inputs={
                'in_file': ('preproc', nifti_gz_format)},
            outputs={
                'moco': ('out_file', nifti_gz_format),
                'moco_par': ('par_file', par_format)},
            requirements=[fsl_req.v('5.0.9')])

        pipeline.add(
            'merge',
            MergeListMotionMat(),
            inputs={
                'file_list': (mcflirt, 'mat_file')},
            outputs={
                'align_mats': ('out_dir', motion_mats_format)})

        return pipeline

    def field_map_time_info_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='field_map_time_info_pipeline',
            desc=("Pipeline to extract delta TE from field map "
                  "images, if provided"),
            citations=[fsl_cite],
            name_maps=name_maps)

        pipeline.add(
            'extract_delta_te',
            FieldMapTimeInfo(),
            inputs={
                'fm_mag': ('field_map_mag', dicom_format)},
            outputs={
                'field_map_delta_te': ('delta_te', float)})

        return pipeline

    def preprocess_pipeline(self, **name_maps):

        if ('field_map_phase' in self.input_names and
                'field_map_mag' in self.input_names):
            return self._fugue_pipeline(**name_maps)
        elif 'reverse_phase' in self.input_names:
            return self._topup_pipeline(**name_maps)
        else:
            return super(EpiStudy, self).preprocess_pipeline(**name_maps)

    def _topup_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='preprocess_pipeline',
            desc=("Topup distortion correction pipeline"),
            citations=[fsl_cite],
            name_maps=name_maps)

        reorient_epi_in = pipeline.add(
            'reorient_epi_in',
            fsl.utils.Reorient2Std(),
            inputs={
                'in_file': ('magnitude', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.9')])

        reorient_epi_opposite = pipeline.add(
            'reorient_epi_opposite',
            fsl.utils.Reorient2Std(),
            inputs={
                'in_file': ('reverse_phase', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.9')])

        prep_dwi = pipeline.add(
            'prepare_dwi',
            PrepareDWI(
                topup=True),
            inputs={
                'pe_dir': ('ped', str),
                'ped_polarity': ('pe_angle', str),
                'dwi': (reorient_epi_in, 'out_file'),
                'dwi1': (reorient_epi_opposite, 'out_file')})

        ped = pipeline.add(
            'gen_config',
            GenTopupConfigFiles(),
            inputs={
                'ped': (prep_dwi, 'pe')})

        merge_outputs = pipeline.add(
            'merge_files',
            merge_lists(2),
            inputs={
                'in1': (prep_dwi, 'main'),
                'in2': (prep_dwi, 'secondary')})

        merge = pipeline.add(
            'fsl_merge',
            fsl_merge(
                dimension='t'),
            inputs={
                'in_files': (merge_outputs, 'out')},
            requirements=[fsl_req.v('5.0.9')])

        topup = pipeline.add(
            'topup',
            TOPUP(),
            inputs={
                'in_file': (merge, 'merged_file'),
                'encoding_file': (ped, 'config_file')},
            requirements=[fsl_req.v('5.0.9')])

        in_apply_tp = pipeline.add(
            'in_apply_tp',
            merge_lists(1),
            inputs={
                'in1': (reorient_epi_in, 'out_file')})

        pipeline.add(
            'applytopup',
            ApplyTOPUP(
                method='jac',
                in_index=[1]),
            inputs={
                'in_files': (in_apply_tp, 'out'),
                'encoding_file': (ped, 'apply_topup_config'),
                'in_topup_movpar': (topup, 'out_movpar'),
                'in_topup_fieldcoef': (topup, 'out_fieldcoef')},
            outputs={
                'preproc': ('out_corrected', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.9')])

        return pipeline

    def _fugue_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='preprocess_pipeline',
            desc=("Fugue distortion correction pipeline"),
            citations=[fsl_cite],
            name_maps=name_maps)

        reorient_epi_in = pipeline.add(
            'reorient_epi_in',
            fsl.utils.Reorient2Std(),
            inputs={
                'in_file': ('magnitude', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.9')])

        fm_mag_reorient = pipeline.add(
            'reorient_fm_mag',
            fsl.utils.Reorient2Std(),
            inputs={
                'in_file': ('field_map_mag', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.9')])

        fm_phase_reorient = pipeline.add(
            'reorient_fm_phase',
            fsl.utils.Reorient2Std(),
            inputs={
                'in_file': ('field_map_phase', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.9')])

        bet = pipeline.add(
            "bet",
            BET(
                robust=True),
            inputs={
                'in_file': (fm_mag_reorient, 'out_file')},
            wall_time=5,
            requirements=[fsl_req.v('5.0.9')])

        create_fmap = pipeline.add(
            "prepfmap",
            PrepareFieldmap(
                # delta_TE=2.46
            ),
            inputs={
                'delta_TE': ('field_map_delta_te', float),
                "in_magnitude": (bet, "out_file"),
                'in_phase': (fm_phase_reorient, 'out_file')},
            wall_time=5,
            requirements=[fsl_req.v('5.0.9')])

        pipeline.add(
            'fugue',
            FUGUE(
                unwarp_direction='x',
                dwell_time=self.parameter('fugue_echo_spacing'),
                unwarped_file='example_func.nii.gz'),
            inputs={
                'fmap_in_file': (create_fmap, 'out_fieldmap'),
                'in_file': (reorient_epi_in, 'out_file')},
            outputs={
                'preproc': ('unwarped_file', nifti_gz_format)},
            wall_time=5,
            requirements=[fsl_req.v('5.0.9')])

        return pipeline

    def motion_mat_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='motion_mat_calculation',
            desc=("Motion matrices calculation"),
            citations=[fsl_cite],
            name_maps=name_maps)

        mm = pipeline.add(
            'motion_mats',
            MotionMatCalculation(),
            inputs={
                'reg_mat': ('coreg_matrix', text_matrix_format),
                'qform_mat': ('qform_mat', text_matrix_format)},
            outputs={
                'motion_mats': ('motion_mats', motion_mats_format)})
        if 'reverse_phase' not in self.input_names:
            pipeline.connect_input('align_mats', mm, 'align_mats',
                                   motion_mats_format)

        return pipeline



import os.path
import errno
import math
from nipype.interfaces.base import (
    File, Directory, TraitedSpec, isdefined, traits, InputMultiPath)
from nipype.interfaces.mrtrix3.reconst import (
    MRTrix3Base, MRTrix3BaseInputSpec)
from arcana.utils import split_extension


# TODO: Write MRtrixBaseInputSpec with all the generic parameters included

# =============================================================================
# Extract MR gradients
# =============================================================================

class DWIPreprocInputSpec(MRTrix3BaseInputSpec):
    in_file = traits.File(
        mandatory=True, argstr='%s',
        desc=("The input DWI series to be corrected"), position=1)
    out_file = File(
        genfile=True, argstr='%s', position=2, hash_files=False,
        desc="Output preprocessed filename")
    out_file_ext = traits.Str(
        desc='Specify the extention for the final output')
    rpe_none = traits.Bool(
        mandatory=False, argstr='-rpe_none',
        desc=("No reverse phase encoded reference image provided"))
    rpe_pair = traits.Bool(
        mandatory=False, argstr="-rpe_pair",
        desc=("Provide a pair of images to use for "
              "inhomogeneity field estimation; note that the FIRST of these "
              "two images must have the same phase"))
    rpe_header = traits.Bool(
        mandatory=False, argstr="-rpe_header",
        desc=(
            "Attempt to read the phase-encoding information from headeer"))
    # Arguments
    pe_dir = traits.Str(
        argstr='-pe_dir %s', desc=(
            "The phase encode direction; can be a signed axis "
            "number (e.g. -0, 1, +2) or a code (e.g. AP, LR, IS)"))
    se_epi = traits.File(
        argstr='-se_epi %s',
        desc=("forward and reverse pair concanenated into a single 4D image "
              "for inhomogeneity field estimation; note that the FIRST of "
              "these two images must have the same phase"),
        position=3)
    eddy_parameters = traits.Str(
        argstr='-eddy_options "%s"', desc='parameters to be passed to eddy')
    no_clean_up = traits.Bool(True, argstr='-nocleanup',
                              desc='Do not delete the temporary folder')
    temp_dir = Directory(genfile=True, argstr='-tempdir %s',
                         desc="Specify the temporay directory")


class DWIPreprocOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='Pre-processed DWI dataset')
    eddy_parameters = File(desc='File with eddy parameters')


class DWIPreproc(MRTrix3Base):

    _cmd = 'dwipreproc'
    input_spec = DWIPreprocInputSpec
    output_spec = DWIPreprocOutputSpec

    def _list_outputs(self):

        outputs = self.output_spec().get()
        outputs['out_file'] = os.path.abspath(self._gen_outfilename())
        outputs['eddy_parameters'] = os.path.abspath(os.path.join(
            self._gen_tempdir(), 'dwi_post_eddy.eddy_parameters'))
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            gen_name = self._gen_outfilename()
        elif name == 'temp_dir':
            gen_name = self._gen_tempdir()
        else:
            assert False
        return gen_name

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            out_name = self.inputs.out_file
        else:
            base, ext = split_extension(
                os.path.basename(self.inputs.in_file))
            if isdefined(self.inputs.out_file_ext):
                extension = self.inputs.out_file_ext
            else:
                extension = ext
            out_name = "{}_preproc{}".format(base, extension)
        return out_name

    def _gen_tempdir(self):
        if isdefined(self.inputs.temp_dir):
            temp_dir = self.inputs.temp_dir
        else:
            base, _ = split_extension(
                os.path.basename(self.inputs.in_file))
            temp_dir = "{}_tempdir".format(base)
        return temp_dir


class DWI2MaskInputSpec(MRTrix3BaseInputSpec):
    # Arguments
    bvalue_scaling = File(
        mandatory=False, argstr='-bvalue_scaling %s',
        desc=("specifies whether the b-values should be scaled by the square "
              "of the corresponding DW gradient norm, as often required for "
              "multi-shell or DSI DW acquisition schemes. The default action "
              "can also be set in the MRtrix config file, under the "
              "BValueScaling entry. Valid choices are yes/no, true/false, "
              "0/1 (default: true)."))
    in_file = traits.File(
        mandatory=True, argstr='%s', exists=True,
        desc=("The input DWI series to be corrected"), position=-2)
    out_file = File(
        genfile=True, argstr='%s', position=-1, hash_files=False,
        desc="Output preprocessed filename")


class DWI2MaskOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='Bias-corrected DWI dataset')


class DWI2Mask(MRTrix3Base):

    _cmd = 'dwi2mask'
    input_spec = DWI2MaskInputSpec
    output_spec = DWI2MaskOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            gen_name = self._gen_outfilename()
        else:
            assert False
        return gen_name

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            out_name = self.inputs.out_file
        else:
            base, ext = split_extension(
                os.path.basename(self.inputs.in_file))
            out_name = os.path.join(
                os.getcwd(), "{}_biascorrect{}".format(base, ext))
        return out_name


class DWIBiasCorrectInputSpec(MRTrix3BaseInputSpec):
    # Arguments
    mask = File(
        mandatory=True, argstr='-mask %s',
        desc=("Whole brain mask"))
    method = traits.Str(
        mandatory=False, argstr='-%s',
        desc=("Method used to correct for biases (either 'fsl' or 'ants')"))
    bias = File(
        mandatory=False, argstr='-bias %s',
        desc=("Output the estimated bias field"))
    in_file = traits.Str(
        mandatory=True, argstr='%s',
        desc=("The input DWI series to be corrected"), position=-2)
    out_file = File(
        genfile=True, argstr='%s', position=-1, hash_files=False,
        desc="Output preprocessed filename")


class DWIBiasCorrectOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='Bias-corrected DWI dataset')


class DWIBiasCorrect(MRTrix3Base):

    _cmd = 'dwibiascorrect'
    input_spec = DWIBiasCorrectInputSpec
    output_spec = DWIBiasCorrectOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            gen_name = self._gen_outfilename()
        else:
            assert False
        return gen_name

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            out_name = self.inputs.out_file
        else:
            base, ext = split_extension(
                os.path.basename(self.inputs.in_file))
            out_name = os.path.join(
                os.getcwd(), "{}_biascorrect{}".format(base, ext))
        return out_name

# =============================================================================
# DWI Denoise
# =============================================================================


class DWIDenoiseInputSpec(MRTrix3BaseInputSpec):
    in_file = traits.Either(
        File(exists=True, desc="Input file"),
        Directory(exists=True, desc="Input directory (assumed to be DICOM)"),
        mandatory=True, argstr='%s', position=-2)
    out_file = File(
        genfile=True, argstr='%s', position=-1, hash_files=False,
        desc=("Output (converted) file. If no path separators (i.e. '/' on "
              "*nix) are found in the provided output file then the CWD (when "
              "the workflow is run, i.e. the working directory) will be "
              "prepended to the output path."))
    out_file_ext = traits.Str(desc='Extention of the output file.')
    noise = File(
        genfile=True, argstr="-noise %s",
        desc=("The estimated spatially-varying noise level"))
    mask = File(
        argstr="-mask %s",
        desc=("Perform the de-noising in the specified mask"))
    extent = traits.Tuple(
        traits.Int(), traits.Int(), traits.Int(), argstr="-extent %d,%d,%d",
        desc="Extent of the kernel")


class DWIDenoiseOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='De noised image')
    noise = File(desc=("The estimated spatially-varying noise level"))


class DWIDenoise(MRTrix3Base):

    _cmd = 'dwidenoise'
    input_spec = DWIDenoiseInputSpec
    output_spec = DWIDenoiseOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfname()
        outputs['noise'] = self._gen_noisefname()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            gen_name = self._gen_outfname()
        elif name == 'noise':
            gen_name = self._gen_noisefname()
        else:
            assert False
        return gen_name

    def _gen_outfname(self):
        if isdefined(self.inputs.out_file):
            out_name = self.inputs.out_file
        else:
            if isdefined(self.inputs.out_file_ext):
                ext = self.inputs.out_file_ext
                base, _ = split_extension(os.path.basename(
                    self.inputs.in_file))
            else:
                base, ext = split_extension(
                    os.path.basename(self.inputs.in_file))
            out_name = os.path.join(os.getcwd(),
                                    "{}_conv{}".format(base, ext))
        return out_name

    def _gen_noisefname(self):
        if isdefined(self.inputs.out_file):
            out_name = self.inputs.out_file
        else:
            if isdefined(self.inputs.out_file_ext):
                ext = self.inputs.out_file_ext
                base, _ = split_extension(os.path.basename(
                    self.inputs.in_file))
            else:
                base, ext = split_extension(
                    os.path.basename(self.inputs.in_file))
            out_name = os.path.join(os.getcwd(),
                                    "{}_noise{}".format(base, ext))
        return out_name


class DWIIntensityNormInputSpec(MRTrix3BaseInputSpec):

    in_files = InputMultiPath(
        File(exists=True),
        desc="The input DWI images to normalize",
        mandatory=True)

    masks = InputMultiPath(
        File(exists=True),
        desc=("Input directory containing brain masks, corresponding to "
              "one per input image in the same order"),
        mandatory=True)

    fa_threshold = traits.Float(
        argstr='-fa_threshold %s',
        desc=("The threshold applied to the Fractional Anisotropy group "
              "template used to derive an approximate white matter mask"))

    fa_template = File(
        genfile=True, hash_files=False,
        desc=("The output population specific FA template, which is "
              "threshold to estimate a white matter mask"),
        argstr='%s', position=-2)

    wm_mask = File(
        genfile=True, hash_files=False,
        desc=("Input directory containing brain masks, corresponding to "
              "one per input image (with the same file name prefix"),
        argstr='%s', position=-1)

    in_dir = Directory(
        genfile=True,
        desc="The input directory to collate the DWI images within",
        argstr='%s', position=-5)

    mask_dir = File(
        genfile=True,
        desc=("Input directory to collate the brain masks within"),
        argstr='%s', position=-4)

    out_dir = Directory(
        genfile=True, argstr='%s', position=-3, hash_files=False,
        desc=("The output directory to containing the normalised DWI images"))


class DWIIntensityNormOutputSpec(TraitedSpec):

    out_files = traits.List(
        File(exists=True),
        desc=("The intensity normalised DWI images"))

    fa_template = File(
        exists=True,
        desc=("The output population specific FA templates, which is "
              "threshold to estimate a white matter mask"))

    wm_mask = File(
        exists=True,
        desc=("Input directory containing brain masks, corresponding to "
              "one per input image (with the same file name prefix"))


class DWIIntensityNorm(MRTrix3Base):

    _cmd = 'dwiintensitynorm'
    input_spec = DWIIntensityNormInputSpec
    output_spec = DWIIntensityNormOutputSpec

    def _run_interface(self, *args, **kwargs):
        self._link_into_dir(self.inputs.in_files, self._gen_in_dir_name())
        self._link_into_dir(self.inputs.masks, self._gen_mask_dir_name())
        return super(DWIIntensityNorm, self)._run_interface(*args, **kwargs)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        out_dir = self._gen_out_dir_name()
        outputs['out_files'] = sorted(
            os.path.join(out_dir, f) for f in os.listdir(out_dir))
        outputs['fa_template'] = self._gen_fa_template_name()
        outputs['wm_mask'] = self._gen_wm_mask_name()
        return outputs

    def _gen_filename(self, name):
        if name == 'in_dir':
            gen_name = self._gen_in_dir_name()
        elif name == 'mask_dir':
            gen_name = self._gen_mask_dir_name()
        elif name == 'out_dir':
            gen_name = self._gen_out_dir_name()
        elif name == 'fa_template':
            gen_name = self._gen_fa_template_name()
        elif name == 'wm_mask':
            gen_name = self._gen_wm_mask_name()
        else:
            assert False
        return gen_name

    def _gen_in_dir_name(self):
        if isdefined(self.inputs.in_dir):
            out_name = self.inputs.in_dir
        else:
            out_name = os.path.join(os.getcwd(), 'in_dir')
        return out_name

    def _gen_mask_dir_name(self):
        if isdefined(self.inputs.mask_dir):
            out_name = self.inputs.mask_dir
        else:
            out_name = os.path.join(os.getcwd(), 'mask_dir')
        return out_name

    def _gen_out_dir_name(self):
        if isdefined(self.inputs.out_dir):
            out_name = self.inputs.out_dir
        else:
            out_name = os.path.join(os.getcwd(), 'out_dir')
        return out_name

    def _gen_fa_template_name(self):
        if isdefined(self.inputs.fa_template):
            out_name = self.inputs.fa_template
        else:
            ext = split_extension(self.inputs.in_files[0])[1]
            out_name = os.path.join(os.getcwd(), 'fa_template' + ext)
        return out_name

    def _gen_wm_mask_name(self):
        if isdefined(self.inputs.wm_mask):
            out_name = self.inputs.wm_mask
        else:
            ext = split_extension(self.inputs.in_files[0])[1]
            out_name = os.path.join(os.getcwd(), 'wm_mask' + ext)
        return out_name

    def _link_into_dir(self, fpaths, dirpath):
        """
        Symlinks the given file paths into the given directory, making the
        directory if necessary
        """
        try:
            os.makedirs(dirpath)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        num_digits = int(math.ceil(math.log(len(fpaths), 10)))
        for i, fpath in enumerate(fpaths):
            _, ext = split_extension(fpath)
            os.symlink(fpath,
                       os.path.join(dirpath, str(i).zfill(num_digits) + ext))



import os.path as op
from nipype.interfaces.matlab import MatlabCommand, MatlabInputSpec
from nipype.interfaces.base import TraitedSpec, traits, File

from banana.interfaces import MATLAB_RESOURCES


class BaseVeinInputSpec(MatlabInputSpec):

    # Need to override value in input spec to make it non-mandatory
    script = traits.Str(
        argstr='-r \"%s;exit\"',
        desc='m-code to run',
        position=-1)


class BaseVeinOutputSpec(TraitedSpec):

    out_file = File(exists=True, desc="Processed file")
    raw_output = traits.Str("Raw output of the matlab command")


class BaseVein(MatlabCommand):
    """
    Base class for MATLAB mask interfaces
    """
    output_spec = BaseVeinOutputSpec

    def run(self, **inputs):
        self.work_dir = inputs['cwd']
        # Set the script input of the matlab spec
        self.inputs.script = (
            "set_param(0,'CharacterEncoding','UTF-8');\n"
            "addpath(genpath('{}'));\n".format(MATLAB_RESOURCES) +
            self.script(**inputs) +
            "exit;")
        results = super().run(**inputs)
        stdout = results.runtime.stdout
        # Attach stdout to outputs to access matlab results
        results.outputs.raw_output = stdout
        return results

    def script(self, **inputs):  # @UnusedVariable
        """
        Generate script to perform masking
        """
        raise NotImplementedError

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self.out_file
        return outputs

    @property
    def out_file(self):
        return op.realpath(op.abspath(op.join(self.work_dir, 'out_file.nii')))


class ShMRFInputSpec(BaseVeinInputSpec):
    in_file = File(exists=True, mandatory=True)
    mask = File(exists=True, mandatory=True)
    omega1 = traits.Float(0.01, usedefault=True)
    omega2 = traits.Float(0.20, usedefault=True)


class ShMRF(BaseVein):
    input_spec = ShMRFInputSpec

    def script(self, **inputs):  # @UnusedVariable
        """
        Generate script to load images, pass them to the STI function along
        with the keyword parameters
        """
        script = """
            % Load input files
            cv_image = load_untouch_nii('{in_file}');
            mask = load_untouch_nii('{mask}');
            out = cv_image;

            mask = mask.img>0;
            cv_image = cv_image.img;
            cv_image(isnan(cv_image)) = 0;

            %%
            params = ShMRF_DefaultParams();
            params.omega1 = {omega1};
            params.omega2 = {omega2}; %(0.12 in first round)
            params.display = false;
            params.preprocess = false;

            out.img = ShMRF_Segment(cv_image, mask, params);

            save_untouch_nii(out, '{out_file}');
        """.format(
            in_file=self.inputs.in_file,
            mask=self.inputs.mask,
            omega1=self.inputs.omega1,
            omega2=self.inputs.omega2,
            out_file=self.out_file)
        return script


class CompositeVeinImageInputSpec(BaseVeinInputSpec):
    qsm = File(exists=True, mandatory=True)
    swi = File(exists=True, mandatory=True)
    mask = File(exists=True, mandatory=True)
    vein_atlas = File(exists=True, mandatory=True)
    q_prior = File(exists=True, mandatory=True)
    s_prior = File(exists=True, mandatory=True)
    a_prior = File(exists=True, mandatory=True)


class CompositeVeinImage(BaseVein):

    input_spec = CompositeVeinImageInputSpec

    def script(self, **inputs):  # @UnusedVariable
        script = """
        mask = load_untouch_nii('{mask}');
        mask = single(mask.img)>0;

        swi = load_untouch_nii('{swi}');
        qsm = load_untouch_nii('{qsm}');
        fre = load_untouch_nii('{vein_atlas}');

        hdrInfo = qsm.hdr;
        hdrInfo.dime.datatype = 64;
        hdrInfo.dime.bitpix = 64;

        swi = single(swi.img);
        qsm = single(qsm.img);
        fre = single(fre.img);

        [ swi, qsm ] = GMM( mask, swi, qsm);
        vein_atlas = min(0.99,max(0.01,fre));

        s_prior = load_untouch_nii('{s_prior}');
        q_prior = load_untouch_nii('{q_prior}');
        a_prior = load_untouch_nii('{a_prior}');

        s_prior = single(s_prior.img);
        q_prior = single(q_prior.img);
        a_prior = single(a_prior.img);

        cvVol = swi.*s_prior + qsm.*q_prior + vein_atlas.*a_prior;
        cvVol = cvVol./(s_prior+q_prior+a_prior);

        cvNii = make_nii(cvVol.*mask);
        cvNii.hdr = hdrInfo;
        save_nii(cvNii, '{out_file}');
        """.format(
            qsm=self.inputs.qsm,
            swi=self.inputs.swi,
            vein_atlas=self.inputs.vein_atlas,
            mask=self.inputs.mask,
            q_prior=self.inputs.q_prior,
            s_prior=self.inputs.s_prior,
            a_prior=self.inputs.a_prior,
            out_file=self.out_file)
        return script



from arcana.citation import Citation


mrtrix_cite = Citation(
    short_name="mrtrix",
    authors=["Tournier, J-D"],
    title="MRtrix Package",
    institute="Brain Research Institute, Melbourne, Australia",
    url="https://github.com/MRtrix3/mrtrix3",
    year=2012)

matlab_cite = Citation(
    short_name="matlab",
    authors=["A"],
    title="MATLAB",
    institute="Mathworks",
    url="https://placeholder.com",
    year=2014)

sti_cites = Citation(
    short_name="sti",
    authors=["A"],
    title="Placeholder",
    institute="Placeholder",
    url="https://placeholder.com",
    year=2014)

eddy_cite = Citation(
    short_name='Eddy',
    authors=["Andersson, J. L.", "Sotiropoulos, S. N."],
    title=(
        "An integrated approach to correction for "
        "off-resonance effects and subject movement in "
        "diffusion MR imaging"),
    journal="NeuroImage", year=2015, volume=125,
    pages=(1063, 1078)),

fsl_cite = Citation(
    short_name="FSL",
    authors=["Smith, S. M.", "Jenkinson, M.",
             "Woolrich, M. W.", "Beckmann, C. F.",
             "Behrens, T. E.", "Johansen- Berg, H.",
             "Bannister, P. R.", "De Luca, M.", "Drobnjak, I.",
             "Flitney, D. E.", "Niazy, R. K.", "Saunders, J.",
             "Vickers, J.", "Zhang, Y.", "De Stefano, N.",
             "Brady, J. M. & Matthews, P. M."],
    title=(
        "Advances in functional and structural MR image "
        "analysis and implementation as FSL"),
    journal="NeuroImage", year=2004, volume=23,
    pages=("S208", "S219"))

distort_correct_cite = Citation(
    short_name="skare_distort_correct",
    authors=["Skare, S.", "Bammer, R."],
    title=(
        "Jacobian weighting of distortion corrected EPI data"),
    journal=(
        "Proceedings of the International Society for Magnetic"
        " Resonance in Medicine"), year=2010, pages=(5063,))

topup_cite = Citation(
    short_name="Topup",
    authors=["Andersson, J. L.", "Skare, S. & Ashburner, J."],
    title=(
        "How to correct susceptibility distortions in "
        "spin-echo echo-planar images: application to "
        "diffusion tensor imaging"),
    journal="NeuroImage", year=2003, volume=20,
    pages=(870, 888))

noddi_cite = Citation(
    short_name="noddi",
    authors=["Zhang, H.", "Schneider T.", "Wheeler-Kingshott C. A.",
             "Alexander, D. C."],
    title=(
        "NODDI: Practical in vivo neurite orientation dispersion and density "
        "imaging of the human brain"),
    journal='NeuroImage', year=2012, volume=61, pages=(1000, 1016))

bet_cite = Citation(
    short_name="BET",
    authors=['Smith, S. M.'],
    title=("Fast robust automated brain extraction"),
    journal="Human Brain Mapping",
    volume=17, issue=3, pages=(143, 155), month='November', year=2002)

bet2_cite = Citation(
    short_name="BET2",
    authors=["Jenkinson, M.", "Pechaud, M.", 'Smith, S. M.'],
    title=("MR-based estimation of brain, skull and scalp surfaces"),
    proceedings=(
        "Eleventh Annual Meeting of the Organization for Human Brain Mapping"),
    year=2005)

fast_cite = Citation(
    short_name="FAST",
    authors=["Zhang, Y.", "Brady, M.", "Smith, S."],
    title=("Segmentation of brain MR images through a hidden Markov random "
           "field model and the expectation-maximization algorithm"),
    proceedings=("IEEE Transactions on Medical Imaging"),
    year=2001, volume=20, pages='45-57')

n4_cite = Citation(
    short_name='n4',
    authors=["Tustison, N.", " Avants, B.", " Cook, P.", " Zheng, Y.",
             " Egan, A.", " Yushkevich, P.", "Gee, J."],
    title=("N4ITK: Improved N3 Bias Correction"),
    proceedings=("IEEE Transactions on Medical Imaging"), year=2010, volume=29,
    pages='1310-1320')


tbss_cite = Citation(
    short_name='tbss',
    authors=['Smith, S.M.', 'Jenkinson, M.', 'Johansen-Berg, H.',
             'Rueckert, D.', 'Nichols, T.E.', 'Mackay, C.E.', 'Watkins, K.E.',
             'Ciccarelli, O.', 'Cader, M.Z.', 'Matthews, P.M.',
             'Behrens, T.E.J.'],
    title=("Tract-based spatial statistics: Voxelwise analysis of "
           "multi-subject diffusion data"),
    journal="NeuroImage", volume=31, pages='1487-1505', year=2006)

spm_cite = Citation(
    short_name='spm',
    authors=['Penny, W.', 'Friston, K.', 'Ashburner, J.', 'Kiebel, S.',
             'Nichols, T.'],
    title=("Statistical Parametric Mapping: The Analysis of Functional Brain "
           "Images"),
    year=2006)

optimal_t1_bet_params_cite = Citation(
    short_name='optimal_t1_bet_params',
    authors=[
        'Popescu V.', 'Battaglini M.', 'Hoogstrate W.S.', 'Verfaillie S.C.',
        'Sluimer I.C.', 'van Schijndel R.A.', 'van Dijk B.W.', 'Cover K.S.',
        'Knol D.L.', 'Jenkinson M.', 'Barkhof F.', 'de Stefano N.',
        'Vrenken H.'],
    title=("Optimizing parameter choice for FSL-Brain Extraction Tool (BET) "
           "on 3D T1 images in multiple sclerosis."),
    journal='Neuroimage', volume=61, year=2012, issue=4)

dwidenoise_cites = [
    Citation(
        short_name='dewidenoise1',
        authors=['Veraart, J.', 'Fieremans. E.', 'Novikov, D.S.'],
        title='Diffusion MRI noise mapping using random matrix theory',
        journal='Magn. Res. Med.', volume=76, issue=5, pages='1582-1593',
        year=2016, doi='10.1002/mrm.26059'),
    Citation(
        short_name='dewidenoise2',
        authors=['Veraart, J.', 'Novikov, D.S.', 'Christiaens, D.',
                 'Adesaron, B.', 'Sijbers, J.', 'Fieremans. E.'],
        title='Denoising of diffusion MRI using random matrix theory',
        journal='NeuroImage', volume=142, pages='394-406',
        year=2016, doi='10.1016/j.neuroimage.2016.08.016')]

freesurfer_cites = [
    Citation(
        short_name='freesurfer1',
        authors=['Dale, A.M.', 'Fischl, B.', 'Sereno, M.I.'],
        title=("Cortical surface-based analysis. I. Segmentation and surface "
               "reconstruction"),
        journal='Neuroimage', volume=9,
        pages='179-194', year=1999),
    Citation(
        short_name='freesurfer2',
        authors=['Dale, A.M.', 'Sereno, M.I.'],
        title=("Improved localization of cortical activity by combining EEG "
               "and MEG with MRI cortical surface reconstruction: a linear "
               "approach"),
        journal='J Cogn Neurosci', volume=5,
        pages='162-176', year=1993),
    Citation(
        short_name='freesurfer3',
        authors=['Desikan, R.S.', 'Segonne, F.', 'Fischl, B.', 'Quinn, B.T.',
                 'Dickerson, B.C.', 'Blacker, D.', 'Buckner, R.L.',
                 'Dale, A.M.', 'Maguire, R.P.', 'Hyman, B.T.', 'Albert, M.S.',
                 'Killiany, R.J.'],
        title=("An automated labeling system for subdividing the human "
               "cerebral cortex on MRI scans into gyral based regions "
               "of interest"),
        journal='Neuroimage', volume=31,
        pages='968-980', year=2006),
    Citation(
        short_name='freesurfer4',
        authors=['Fischl, B.', 'Dale, A.M.'],
        title=("Measuring the thickness of the human cerebral cortex from "
               "magnetic resonance images"),
        journal='Proc Natl Acad Sci USA', volume=97,
        pages='11050-11055', year=2000),
    Citation(
        short_name='freesurfer5',
        authors=['Fischl, B.', 'Liu, A.', 'Dale, A.M.'],
        title=("Automated manifold surgery: constructing geometrically "
               "accurate and topologically correct models of the human "
               "cerebral cortex"),
        journal='IEEE Trans Med Imaging', volume=20,
        pages='70-80', year=2001),
    Citation(
        short_name='freesurfer6',
        authors=['Fischl, B.', 'Salat, D.H.', 'Busa, E.', 'Albert, M.',
                 'Dieterich, M.', 'Haselgrove, C.', 'van der Kouwe, A.',
                 'Killiany, R.', 'Kennedy, D.', 'Klaveness, S.',
                 'Montillo, A.', 'Makris, N.', 'Rosen, B.', 'Dale, A.M.'],
        title=("Whole brain segmentation: automated labeling of "
               "neuroanatomical structures in the human brain"),
        journal='Neuron', volume=33,
        pages='341-355', year=2002),
    Citation(
        short_name='freesurfer7',
        authors=['Fischl, B.', 'Salat, D.H.', 'van der Kouwe, A.J.',
                 'Makris, N.', 'Segonne, F.', 'Quinn, B.T.', 'Dale, A.M.'],
        title=("Sequence-independent segmentation of magnetic resonance "
               "images"),
        journal='Neuroimage 23 Suppl', volume=1,
        pages='S69-84', year=2004),
    Citation(
        short_name='freesurfer8',
        authors=['Fischl, B.', 'Sereno, M.I.', 'Dale, A.M.'],
        title=("Cortical surface-based analysis. II: Inflation, flattening, "
               "and a surface-based coordinate system"),
        journal='Neuroimage', volume=9,
        pages='195-207', year=1999),
    Citation(
        short_name='freesurfer9',
        authors=['Fischl, B.', 'Sereno, M.I.', 'Tootell, R.B.', 'Dale, A.M.'],
        title=("High-resolution intersubject averaging and a coordinate "
               "system for the cortical surface"),
        journal='Hum Brain Mapp', volume=8,
        pages='272-284', year=1999),
    Citation(
        short_name='freesurfer10',
        authors=['Fischl, B.', 'van der Kouwe, A.', 'Destrieux, C.',
                 'Halgren, E.', 'Segonne, F.', 'Salat, D.H.', 'Busa, E.',
                 'Seidman, L.J.', 'Goldstein, J.', 'Kennedy, D.',
                 'Caviness, V.', 'Makris, N.', 'Rosen, B.', 'Dale, A.M.'],
        title=("Automatically parcellating the human cerebral cortex"),
        journal='Cereb Cortex', volume=14,
        pages='11-22', year=2004),
    Citation(
        short_name='freesurfer11',
        authors=['Han, X.', 'Jovicich, J.', 'Salat, D.', 'van der Kouwe, A.',
                 'Quinn, B.', 'Czanner, S.', 'Busa, E.', 'Pacheco, J.',
                 'Albert, M.', 'Killiany, R.', 'Maguire, P.', 'Rosas, D.',
                 'Makris, N.', 'Dale, A.', 'Dickerson, B.', 'Fischl, B.'],
        title=("Reliability of MRI-derived measurements of human cerebral "
               "cortical thickness: the effects of field strength, scanner "
               "upgrade and manufacturer"),
        journal='Neuroimage', volume=32,
        pages='180-194', year=2006),
    Citation(
        short_name='freesurfer12',
        authors=['Jovicich, J.', 'Czanner, S.', 'Greve, D.', 'Haley, E.',
                 'van der Kouwe, A.', 'Gollub, R.', 'Kennedy, D.',
                 'Schmitt, F.', 'Brown, G.', 'Macfall, J.', 'Fischl, B.',
                 'Dale, A.'],
        title=("Reliability in multi-site structural MRI studies: effects of "
               "gradient non-linearity correction on phantom and human data"),
        journal='Neuroimage', volume=30,
        pages='436-443', year=2006),
    Citation(
        short_name='freesurfer13',
        authors=['Kuperberg, G.R.', 'Broome, M.R.', 'McGuire, P.K.',
                 'David, A.S.', 'Eddy, M.', 'Ozawa, F.', 'Goff, D.',
                 'West, W.C.', 'Williams, S.C.', 'van der Kouwe, A.J.',
                 'Salat, D.H.', 'Dale, A.M.', 'Fischl, B.'],
        title=("Regionally localized thinning of the cerebral cortex in "
               "schizophrenia"),
        journal='Arch Gen Psychiatry', volume=60,
        pages='878-888', year=2003),
    Citation(
        short_name='freesurfer14',
        authors=['Reuter, M.', 'Schmansky, N.J.', 'Rosas, H.D.', 'Fischl, B.'],
        title=("Within-Subject Template Estimation for Unbiased Longitudinal "
               "Image Analysis"),
        journal='Neuroimage', volume=61, issue=4,
        pages='1402-1418', year=2012,
        pdf='http://reuter.mit.edu/papers/reuter-long12.pdf'),
    Citation(
        short_name='freesurfer15',
        authors=['Reuter, M.', 'Fischl, B.'],
        title=("Avoiding asymmetry-induced bias in longitudinal image "
               "processing"),
        journal='Neuroimage', volume=57, issue=1,
        pages='19-21', year=2011,
        pdf='http://reuter.mit.edu/papers/reuter-bias11.pdf'),
    Citation(
        short_name='freesurfer16',
        authors=['Reuter, M.', 'Rosas, H.D.', 'Fischl, B.'],
        title=("Highly Accurate Inverse Consistent Registration: A Robust "
               "Approach"),
        journal='Neuroimage', volume=53, issue=4,
        pages='1181-1196', year=2010,
        pdf='http://reuter.mit.edu/papers/reuter-robreg10.pdf'),
    Citation(
        short_name='freesurfer17',
        authors=['Rosas, H.D.', 'Liu, A.K.', 'Hersch, S.', 'Glessner, M.',
                 ' Ferrante, R.J.', 'Salat, D.H.', 'van der Kouwe, A.',
                 ' Jenkins, B.G.', 'Dale, A.M.', 'Fischl, B.'],
        title=("Regional and progressive thinning of the cortical ribbon in "
               "Huntington's disease"),
        journal='Neurology', volume=58,
        pages='695-701', year=2002),
    Citation(
        short_name='freesurfer18',
        authors=['Salat, D.H.', 'Buckner, R.L.', 'Snyder, A.Z.',
                 ' Greve, D.N.', 'Desikan, R.S.', 'Busa, E.', 'Morris, J.C.',
                 'Dale, A.M.', 'Fischl, B.'],
        title=("Thinning of the cerebral cortex in aging"),
        journal='Cereb Cortex', volume=14,
        pages='721-730', year=2004),
    Citation(
        short_name='freesurfer19',
        authors=['Segonne, F.', 'Dale, A.M.', 'Busa, E.', 'Glessner, M.',
                 'Salat, D.', 'Hahn, H.K.', 'Fischl, B.'],
        title=("A hybrid approach to the skull stripping problem in MRI"),
        journal='Neuroimage', volume=22,
        pages='1060-1075', year=2004),
    Citation(
        short_name='freesurfer20',
        authors=['Segonne, F.', 'Pacheco, J.', 'Fischl, B.'],
        title=("Geometrically accurate topology-correction of cortical "
               "surfaces using nonseparating loops"),
        journal='IEEE Trans Med Imaging', volume=26,
        pages='518-529', year=2007),
    Citation(
        short_name='freesurfer21',
        authors=['Sled, J.G.', 'Zijdenbos, A.P.', 'Evans, A.C.'],
        title=("A nonparametric method for automatic correction of intensity "
               "nonuniformity in MRI data"),
        journal='IEEE Trans Med Imaging', volume=17,
        pages='87-97', year=1998)]




from nipype.interfaces.base import (BaseInterface, BaseInterfaceInputSpec,
                                    traits, File, TraitedSpec)
import nibabel as nib
import numpy as np
from sklearn.decomposition import FastICA as fICA
from nipype.utils.filemanip import split_filename
import os


class FastICAInputSpec(BaseInterfaceInputSpec):

    volume = File(exists=True, desc='4D file to be decomposed using ICA',
                  mandatory=True)
    n_components = traits.Int(desc='Number of ICA components to extract',
                              mandatory=True)
    ica_type = traits.Str(desc='Type of ICA to run. Possible types are '
                          'spatial (default) and temporal.', default='spatial')


class FastICAOutputSpec(TraitedSpec):

    ica_decomposition = File(
        exists=True, desc='Nifti file containing ICA decomposition')
    ica_timeseries = File(
        exists=True, desc='Nifti file containing ICA timecourse plots')
    mixing_mat = File(
        exists=True, desc='Text file containing ICA mixing matrix')


class FastICA(BaseInterface):
    input_spec = FastICAInputSpec
    output_spec = FastICAOutputSpec

    def _run_interface(self, runtime):
        fname = self.inputs.volume
        img = nib.load(fname)
        comp = self.inputs.n_components
        data = np.array(img.get_data())

        n_voxels = data.shape[0]*data.shape[1]*data.shape[2]
        ts = data.reshape(n_voxels, data.shape[3])
        _, base, _ = split_filename(fname)

        if self.inputs.ica_type == 'spatial':
            ica_input = ts.T
            outname = 'sICA'
        else:
            ica_input = ts
            outname = 'tICA'

        # Run ICA
        ica = fICA(n_components=comp)
        ica.fit(ica_input)
        S_ = ica.fit_transform(ica_input)
        if self.inputs.ica_type == 'spatial':
            sm = ica.components_.T[:]
            tc = S_[:]
        else:
            sm = S_[:]
            tc = ica.components_.T[:]

        ica_zscore = np.zeros((data.shape[0], data.shape[1],
                               data.shape[2], self.inputs.n_components))
        ica_tc = np.zeros((data.shape[3], self.inputs.n_components))
        for i in range(self.inputs.n_components):
            dt = sm[:, i]-np.mean(sm[:, i])
            num = np.mean(dt**3)
            denom = (np.mean(dt**2))**1.5
            s = num / denom
            print(s)
            if np.sign(s) == -1:
                print('Flipping sign of component {}'.format(str(i)))
                sm[:, i] = -1*sm[:, i]
                tc[:, i] = -1*tc[:, i]

            pc = sm[:, i].reshape(
                data.shape[0], data.shape[1], data.shape[2])

            vstd = np.linalg.norm(sm[:, i])/np.sqrt(n_voxels-1)
            if vstd != 0:
                pc_zscore = pc/vstd
            else:
                print ('Not converting to z-scores as division by zero'
                       ' warning may occur.')
                pc_zscore = pc
            ica_zscore[:, :, :, i] = pc_zscore
            ica_tc[:, i] = tc[:, i]

        im2save = nib.Nifti1Image(ica_zscore, affine=img.affine)
        tc2save = nib.Nifti1Image(ica_tc, affine=np.eye(4))
        nib.save(
            im2save, '{0}_{1}_results_pc{2}_zscore.nii.gz'
            .format(base, outname, str(self.inputs.n_components)))
        nib.save(
            tc2save, '{0}_{1}_timecourse_pc{2}.nii.gz'
            .format(base, outname, str(self.inputs.n_components)))
        np.savetxt(
            '{0}_{1}_mixing_matrix_pc{2}.txt'.format(
                base, outname, str(self.inputs.n_components)), S_)

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        fname = self.inputs.volume
        if self.inputs.ica_type == 'spatial':
            outname = 'sICA'
        else:
            outname = 'tICA'
        _, base, _ = split_filename(fname)
        outputs["ica_decomposition"] = os.path.abspath(
            base+'_{0}_results_pc{1}_zscore.nii.gz'.format(
                outname, str(self.inputs.n_components)))
        outputs["ica_timeseries"] = os.path.abspath(
            base+'_{0}_timecourse_pc{1}.nii.gz'.format(
                outname, str(self.inputs.n_components)))
        outputs["mixing_mat"] = os.path.abspath(
            base+'_{0}_mixing_matrix_pc{1}.txt'.format(
                outname, str(self.inputs.n_components)))

        return outputs



from itertools import groupby, chain
from operator import itemgetter
from nipype.interfaces.matlab import MatlabCommand, MatlabInputSpec
from nipype.interfaces.base import TraitedSpec, traits, File
from banana.exceptions import BananaRuntimeError
import os.path as op
from banana.interfaces import MATLAB_RESOURCES


class BaseSTIInputSpec(MatlabInputSpec):

    # Need to override value in input spec to make it non-mandatory
    script = traits.Str(
        argstr='-r \"%s;exit\"',
        desc='m-code to run',
        position=-1)


class BaseSTIOutputSpec(TraitedSpec):
    raw_output = traits.Str("Raw output of the matlab command")


class BaseSTICommand(MatlabCommand):
    """
    Base interface for STI classes
    """

    def run(self, **inputs):
        # Set the script input of the matlab spec
        self.inputs.script = self.script(cwd=inputs['cwd'])
        results = super().run(**inputs)
        stdout = results.runtime.stdout
        # Attach stdout to outputs to access matlab results
        results.outputs.raw_output = stdout
        return results

    def script(self, cwd, **kwargs):
        """
        Generate script to load images, pass them to the STI function along
        with the keyword parameters
        """
        script = self._set_path()
        script += self._create_param_structs()
        script += self._process_image(cwd, **kwargs)
        script += self._exit()
        return script

    def _process_image(self, cwd, **kwargs):
        script = self._load_images(**kwargs)
        script += self._create_function_call()
        script += self._save_images(cwd=cwd, **kwargs)
        return script

    def _set_path(self):
        script = "set_param(0,'CharacterEncoding','UTF-8');\n"
        # Add path for nifti loader/unloader (might be unecessary
        script += "addpath(genpath('{}'));\n".format(MATLAB_RESOURCES)
        return script

    def _load_images(self, **kwargs):
        "Load all input images"
        script = ''
        for name, _ in self.input_imgs:
            script += "{}_img = load_untouch_nii('{}');\n".format(
                name, self._input_fname(name, **kwargs))
            script += "{name} = {name}_img.img;\n".format(
                name=name)
        return script

    def _input_fname(self, name, **kwargs):  # @UnusedVariable
        return getattr(self.inputs, name)

    def _output_fname(self, name, **kwargs):  # @UnusedVariable
        return name

    def _create_param_structs(self):
        "Create parameter structs"
        script = ''
        for name, struct in self.structs:
            for keyword in struct:
                script += '{}.{} = {};\n'.format(name, keyword,
                                                 getattr(self.inputs, keyword))
        return script

    def _create_function_call(self):
        "Create function call"
        # Get input arguments
        input_args = []
        for name, trait in self.input_imgs:
            if trait.format_str is not None:
                arg = getattr(self.inputs, trait.format_str).format(name)
            else:
                arg = name
            input_args.append(arg)
        # Create output arguments
        output_args = (o for o, _ in self.output_imgs)
        # Core part of function call: output array, function name + inputs
        script = '[{}] = {}({}'.format(
            ', '.join(output_args), self.func, ', '.join(input_args))
        # Append keyword arguments to function call if required
        if self.has_keywords:
            script += ', ' + ', '.join(
                "'{}', {}".format(*kw) for kw in self.keyword_args)
        script += ');\n'
        return script

    def _save_images(self, cwd, **kwargs):
        "Save all output images"
        script = ''
        for name, trait in self.output_imgs:
            # Copy image to pull header from
            script += "{}_img = {}_img;\n".format(
                name, trait.header_from)
            script += "{name}_img.img = {name};\n".format(name=name)
            script += "save_untouch_nii({}_img, '{}.nii');\n".format(
                name, op.join(cwd, self._output_fname(name, **kwargs)))
        return script

    def _exit(self):
        return 'exit;\n'

    @property
    def input_imgs(self):
        """
        Return all input traits with 'argpos' metadata, sorted by argpos
        """
        return sorted(
            ((n, t) for n, t in self.inputs.items() if t.argpos is not None),
            key=lambda x: x[1].argpos)

    @property
    def output_imgs(self):
        return sorted(
            ((n, t) for n, t in self._outputs().items()
             if t.outpos is not None),
            key=lambda x: x[1].outpos)

    @property
    def structs(self):
        key = itemgetter(0)
        for struct, keywords in groupby(sorted(
            ((t.in_struct, (n, t)) for n, t in self.inputs.items()
             if t.in_struct is not None), key=key), key=key):
            yield struct, dict(kw for _, kw in keywords)

    @property
    def has_keywords(self):
        return (any(i.keyword for i in self.inputs.traits().values()) or
                list(self.structs))

    @property
    def keyword_args(self):
        return chain(((n, getattr(self.inputs, n))
                      for n, i in self.inputs.items() if i.keyword),
                     ((n, n) for n, _ in self.structs))

    def _list_outputs(self):
        outputs = self._outputs().get()
        for name, _ in self.output_imgs:
            outputs[name] = op.abspath(self._output_fname(name)) + '.nii.gz'
        return outputs

    def _index_suffix(self, index):
        return str(index)


class UnwrapPhaseInputSpec(BaseSTIInputSpec):

    in_file = File(exists=True, mandatory=True, argpos=0, formatstr="{}",
                   desc="Input file to unwrap")
    voxelsize = traits.List([traits.Float(), traits.Float(), traits.Float()],
                             mandatory=True, keyword=True,
                             desc="Voxel size of the image")
    padsize = traits.List([12, 12, 12],
                          (traits.Int(), traits.Int(), traits.Int()),
                          usedefault=True, keyword=True,
                          desc="Padding size for each dimension")


class UnwrapPhaseOutputSpec(BaseSTIOutputSpec):

    out_file = File(exists=True, outpos=0, desc="Unwrapped phase image",
                    header_from='in_file')


class UnwrapPhase(BaseSTICommand):

    func = 'MRPhaseUnwrap'
    input_spec = UnwrapPhaseInputSpec
    output_spec = UnwrapPhaseOutputSpec


class VSharpInputSpec(BaseSTIInputSpec):

    in_file = File(exists=True, mandatory=True, argpos=0,
                   desc="Input file to unwrap")
    mask = File(exists=True, mandatory=True, argpos=1, desc="Mask file",
                format_str='mask_manip')
    mask_manip = traits.Str(
        desc=("A format string used to manipulate the mask before it is "
              "passed as an argument to the function"))
    voxelsize = traits.List([traits.Float(), traits.Float(), traits.Float()],
                             mandatory=True, keyword=True,
                             desc="Voxel size of the image")


class VSharpOutputSpec(BaseSTIOutputSpec):

    out_file = File(exists=True, outpos=0, desc="Unwrapped phase image",
                    header_from='in_file')
    new_mask = File(exists=True, outpos=1, desc="New mask",
                    header_from='mask')


class VSharp(BaseSTICommand):

    func = 'V_SHARP'
    input_spec = VSharpInputSpec
    output_spec = VSharpOutputSpec


class QSMiLSQRInputSpec(BaseSTIInputSpec):

    in_file = File(exists=True, mandatory=True, argpos=0,
                   desc="Input file to unwrap")
    mask = File(exists=True, mandatory=True, argpos=1,
                desc="Input file to unwrap", format_str='mask_manip')
    mask_manip = traits.Str(
        desc=("The format string used to manipulate the mask before it is "
              "passed as an argument to the function"))
    voxelsize = traits.List([traits.Float(), traits.Float(), traits.Float()],
                             mandatory=True, in_struct='params',
                             desc="Voxel size of the image")
    padsize = traits.List([12, 12, 12],
                          (traits.Int(), traits.Int(), traits.Int()),
                          usedefault=True, in_struct='params',
                          desc="Padding size for each dimension")
    te = traits.Float(mandatory=True, desc="TE time of acquisition protocol",
                      in_struct='params')
    B0 = traits.Float(mandatory=True, desc="B0 field strength",
                      in_struct='params')
    H = traits.List((traits.Float(), traits.Float(), traits.Float()),
                     mandatory=True, desc="Direction of the B0 field",
                     in_struct='params')


class QSMiLSQROutputSpec(BaseSTIOutputSpec):

    out_file = File(exists=True, outpos=0, desc="Unwrapped phase image",
                    header_from='in_file')


class QSMiLSQR(BaseSTICommand):

    func = 'QSM_iLSQR'
    input_spec = QSMiLSQRInputSpec
    output_spec = QSMiLSQROutputSpec


class BaseBatchSTICommand(BaseSTICommand):
    """
    Runs an STI command in batch mode over a range of input files instead of
    a single one.

    Although a MapNode could be used instead, it would incur a penalty of
    opening and closing MATLAB for each iteration of the batch.
    """

    def run(self, **inputs):
        # Determine batch size
        self.batch_size = None
        for img_name, _ in self.input_imgs:
            inpt = getattr(self.inputs, img_name)
            if isinstance(inpt, list):
                if self.batch_size is None:
                    self.batch_size = len(inpt)
                elif self.batch_size != len(inpt):
                    raise BananaRuntimeError(
                        "Inconsistent length of batch input lists ({} and {})"
                        .format(len(inpt), self.batch_size))
        return super(BaseBatchSTICommand, self).run(**inputs)

    def script(self, cwd, **kwargs):
        """
        Generate script to load images, pass them to the STI function along
        with the keyword parameters
        """
        script = self._set_path()
        script += self._create_param_structs()
        for i in range(self.batch_size):
            script += self._process_image(cwd, index=i, **kwargs)
        script += self._exit()
        return script

    def _input_fname(self, name, index, **kwargs):  # @UnusedVariable
        inpt = super(BaseBatchSTICommand, self)._input_fname(name)
        if isinstance(inpt, list):
            inpt = inpt[index]
        return inpt

    def _output_fname(self, name, index, **kwargs):  # @UnusedVariable
        return name + str(index)

    def _list_outputs(self):
        outputs = self._outputs().get()
        for name, _ in self.output_imgs:
            outputs[name] = []
            for i in range(self.batch_size):
                outputs[name].append(op.abspath(self._output_fname(name, i)) +
                                     '.nii')
        return outputs


class BatchUnwrapPhaseInputSpec(UnwrapPhaseInputSpec):

    in_file = traits.List(
        File(exists=True), mandatory=True, argpos=0, formatstr="{}",
        desc="Input file to unwraps")


class BatchUnwrapPhaseOutputSpec(UnwrapPhaseOutputSpec):

    out_file = traits.List(File(exists=True),
                           outpos=0, desc="Unwrapped phase images",
                           header_from='in_file')


class BatchUnwrapPhase(BaseBatchSTICommand, UnwrapPhase):

    input_spec = BatchUnwrapPhaseInputSpec
    output_spec = BatchUnwrapPhaseOutputSpec


class BatchVSharpInputSpec(VSharpInputSpec):

    in_file = traits.List(File(exists=True), mandatory=True, argpos=0,
                          desc="Input files to unwrap")
    mask = traits.List(File(exists=True), mandatory=True, argpos=1,
                       desc="Mask files", format_str='mask_manip')


class BatchVSharpOutputSpec(VSharpOutputSpec):

    out_file = traits.List(
        File(exists=True), outpos=0, desc="Unwrapped phase images",
        header_from='in_file')
    new_mask = traits.List(
        File(exists=True), outpos=1, desc="New masks", header_from='mask')


class BatchVSharp(BaseBatchSTICommand, VSharp):

    input_spec = BatchVSharpInputSpec
    output_spec = BatchVSharpOutputSpec


class BatchQSMiLSQRInputSpec(QSMiLSQRInputSpec):

    in_file = traits.List(
        File(exists=True), mandatory=True, argpos=0,
        desc="Input files to unwrap")
    mask = traits.List(
        File(exists=True), mandatory=True, argpos=1,
        desc="Input files to unwrap", format_str='mask_manip')


class BatchQSMiLSQROutputSpec(QSMiLSQROutputSpec):

    out_file = traits.List(
        File(exists=True), outpos=0,
        desc="Unwrapped phase images", header_from='in_file')


class BatchQSMiLSQR(BaseBatchSTICommand, QSMiLSQR):

    input_spec = BatchQSMiLSQRInputSpec
    output_spec = BatchQSMiLSQROutputSpec




from nipype.interfaces.base import (
    isdefined, CommandLineInputSpec, TraitedSpec, CommandLine, File,
    traits)
import os
from arcana.utils import split_extension


class ANTs2FSLMatrixConversionInputSpec(CommandLineInputSpec):

    itk_file = File(exists=True, argstr="-itk %s", position=1)
    reference_file = File(exists=True, argstr="-ref %s", position=2)
    source_file = File(exists=True, argstr='-src %s', position=3)
    ras2fsl = traits.Bool(argstr='-ras2fsl', position=4)
    fsl_matrix = File(genfile=True, argstr='-o %s', position=5)


class ANTs2FSLMatrixConversionOutputSpec(TraitedSpec):

    fsl_matrix = File(exists=True)


class ANTs2FSLMatrixConversion(CommandLine):

    _cmd = 'c3d_affine_tool'
    input_spec = ANTs2FSLMatrixConversionInputSpec
    output_spec = ANTs2FSLMatrixConversionOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['fsl_matrix'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'fsl_matrix':
            gen_name = self._gen_outfilename()
        else:
            assert False
        return gen_name

    def _gen_outfilename(self):
        if isdefined(self.inputs.fsl_matrix):
            out_name = self.inputs.fsl_matrix
        else:
            base, _ = split_extension(
                os.path.basename(self.inputs.itk_file))
            out_name = os.path.join(os.getcwd(),
                                    "{}_fsl.mat".format(base))
        return out_name



# from scipy import ndimage as ndi
# import nibabel as nib
# 
# img1 = nib.load('my_file.nii')
# img2 = nib.load('other_file.nii.gz')
# img3 = nib.load('spm_file.img')
# 
# data = img1.get_data()
# affine = img1.affine
# 
# print(img1)
# 
# nib.save(img1, 'my_file_copy.nii.gz')
# 
# new_image = nib.Nifti1Image(data, affine)
# nib.save(new_image, 'new_image.nii.gz')




import sys
import os.path
from setuptools import setup, find_packages

PACKAGE_NAME = 'banana'

# Get version from module inside package
sys.path.insert(0, os.path.join(os.path.dirname(__file__), PACKAGE_NAME))
from __about__ import __version__, install_requires  # @UnresolvedImport @IgnorePep8
sys.path.pop(0)


setup(
    name=PACKAGE_NAME,
    version=__version__,
    author='Tom G. Close',
    author_email='tom.g.close@gmail.com',
    packages=find_packages(),
    url='https://github.com/MonashBI/{}'.format(PACKAGE_NAME),
    license='The Apache Software Licence 2.0',
    description=(
        'Biomedical ANAlysis iN Arcana (Banana): biomedical '
        'analysis workflows implemented in the Arcana '
        'framework (arcana.readthedocs.io)'),
    long_description=open('README.rst').read(),
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'gen_banana_test_data = '
            'banana.utils.testing:gen_test_data_entry_point']},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Medical Science Apps."])




from nipype.interfaces.base import (BaseInterface, BaseInterfaceInputSpec,
                                    traits, File, TraitedSpec)
import nibabel as nib
import numpy as np
from nipype.utils.filemanip import split_filename
import os
import matplotlib.pyplot as plot
from sklearn.decomposition import PCA
import subprocess as sp
from nipype.interfaces.base.traits_extension import Directory, isdefined
import shutil
import glob
import pydicom
from nipype.interfaces import fsl


list_mode_framing_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..', 'resources', 'C_C++',
                 'ListModeFraming'))
interfile_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..', 'resources', 'pet',
                 'biograph_mmr_short_int.hs'))


class PETdrInputSpec(BaseInterfaceInputSpec):

    volume = File(exists=True, desc='4D input for the dual regression',
                  mandatory=True)
    regression_map = File(exists=True, desc='3D map to use for the spatial '
                          'regression (first step of the dr)', mandatory=True)
    threshold = traits.Float(desc='Threshold to be applied to the abs(reg_map)'
                             ' before regression (default zero)', default=0.0)
    binarize = traits.Bool(desc='If True, all the voxels greater than '
                           'threshold will be set to 1 (default False)',
                           default=False)


class PETdrOutputSpec(TraitedSpec):

    spatial_map = File(
        exists=True, desc='Nifti file containing result for the temporal '
        'regression')
    timecourse = File(
        exists=True, desc='Png file containing result for the spatial '
        'regression')


class PETdr(BaseInterface):

    input_spec = PETdrInputSpec
    output_spec = PETdrOutputSpec

    def _run_interface(self, runtime):
        fname = self.inputs.volume
        mapname = self.inputs.regression_map
        th = self.inputs.threshold
        binarize = self.inputs.binarize
        _, base, _ = split_filename(fname)
        _, base_map, _ = split_filename(mapname)

        img = nib.load(fname)
        data = np.array(img.get_data())
        spatial_regressor = nib.load(mapname)
        spatial_regressor = np.array(spatial_regressor.get_data())

        n_voxels = data.shape[0]*data.shape[1]*data.shape[2]
        ts = data.reshape(n_voxels, data.shape[3])
        mask = spatial_regressor.reshape(n_voxels, 1)
        if th and not binarize:
            mask[np.abs(mask) < th] = 0
            base = base+'_th_{}'.format(str(th))
        elif th and binarize:
            mask[mask < th] = 0
            mask[mask >= th] = 1
            base = base+'_bin_th_{}'.format(str(th))
        timecourse = np.dot(ts.T, mask)
        sm = np.dot(ts, timecourse)
        mean = np.mean(sm)
        std = np.std(sm)
        sm_zscore = (sm-mean)/std

        im2save = nib.Nifti1Image(
            sm_zscore.reshape(spatial_regressor.shape), affine=img.affine)
        nib.save(
            im2save, '{0}_{1}_GLM_fit_zscore.nii.gz'.format(base, base_map))

        plot.plot(timecourse)
        plot.savefig('{0}_{1}_timecourse.png'.format(base, base_map))
        plot.close()

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        fname = self.inputs.volume
        th = self.inputs.threshold
        binarize = self.inputs.binarize
        mapname = self.inputs.regression_map

        _, base_map, _ = split_filename(mapname)
        _, base, _ = split_filename(fname)
        if th and not binarize:
            base = base+'_th_{}'.format(str(th))
        elif th and binarize:
            base = base+'_bin_th_{}'.format(str(th))

        outputs["spatial_map"] = os.path.abspath(
            '{0}_{1}_GLM_fit_zscore.nii.gz'.format(base, base_map))
        outputs["timecourse"] = os.path.abspath(
            '{0}_{1}_timecourse.png'.format(base, base_map))

        return outputs


class GlobalTrendRemovalInputSpec(BaseInterfaceInputSpec):

    volume = File(exists=True, desc='4D input file',
                  mandatory=True)


class GlobalTrendRemovalOutputSpec(TraitedSpec):

    detrended_file = File(
        exists=True, desc='4D file with the first temporal PCA component'
        ' removed')


class GlobalTrendRemoval(BaseInterface):

    input_spec = GlobalTrendRemovalInputSpec
    output_spec = GlobalTrendRemovalOutputSpec

    def _run_interface(self, runtime):

        fname = self.inputs.volume
        _, base, _ = split_filename(fname)

        img = nib.load(fname)
        data = np.array(img.get_data())

        n_voxels = data.shape[0]*data.shape[1]*data.shape[2]
        ts = data.reshape(n_voxels, data.shape[3])
        pca = PCA(50)
        pca.fit(ts)
        baseline = np.reshape(
            pca.components_[0, :], (len(pca.components_[0, :]), 1))
        new_ts = ts.T-np.dot(baseline, np.dot(np.linalg.pinv(baseline), ts.T))
        im2save = nib.Nifti1Image(
            new_ts.T.reshape(data.shape), affine=img.affine)
        nib.save(
            im2save, '{}_baseline_removed.nii.gz'.format(base))

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        fname = self.inputs.volume

        _, base, _ = split_filename(fname)

        outputs["detrended_file"] = os.path.abspath(
            '{}_baseline_removed.nii.gz'.format(base))

        return outputs


class SUVRCalculationInputSpec(BaseInterfaceInputSpec):

    volume = File(exists=True, desc='3D input file',
                  mandatory=True)
    base_mask = File(exists=True, desc='3D baseline mask',
                     mandatory=True)


class SUVRCalculationOutputSpec(TraitedSpec):

    SUVR_file = File(
        exists=True, desc='3D SUVR file')


class SUVRCalculation(BaseInterface):

    input_spec = SUVRCalculationInputSpec
    output_spec = SUVRCalculationOutputSpec

    def _run_interface(self, runtime):

        fname = self.inputs.volume
        maskname = self.inputs.base_mask
        _, base, _ = split_filename(fname)

        img = nib.load(fname)
        data = np.array(img.get_data())
        mask = nib.load(maskname)
        mask = np.array(mask.get_data())

        [x, y, z] = np.where(mask > 0)
        ii = np.arange(x.shape[0])
        mean_uptake = np.mean(data[x[ii], y[ii], z[ii]])
        new_data = data / mean_uptake
        im2save = nib.Nifti1Image(new_data, affine=img.affine)
        nib.save(im2save, '{}_SUVR.nii.gz'.format(base))

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        fname = self.inputs.volume

        _, base, _ = split_filename(fname)

        outputs["SUVR_file"] = os.path.abspath(
            '{}_SUVR.nii.gz'.format(base))

        return outputs


class PrepareUnlistingInputsInputSpec(BaseInterfaceInputSpec):

    time_offset = traits.Int(desc='Time between the PET start time and the '
                             'time when you want to initiate the sinogram '
                             'sorting (in seconds).')
    num_frames = traits.Int(desc='Number of frame you want to unlist.')
    temporal_len = traits.Float(desc='Temporal duration, in seconds, of each '
                                'frame. Minumum is 0.001.')
    list_mode = File(exists=True, desc='Listmode data')


class PrepareUnlistingInputsOutputSpec(TraitedSpec):

    out = traits.List(desc='List of all the outputs for PETListModeUnlisting '
                      'pipeline')


class PrepareUnlistingInputs(BaseInterface):

    input_spec = PrepareUnlistingInputsInputSpec
    output_spec = PrepareUnlistingInputsOutputSpec

    def _run_interface(self, runtime):

        time_offset = self.inputs.time_offset
        num_frames = self.inputs.num_frames
        temporal_len = self.inputs.temporal_len
        list_mode = self.inputs.list_mode

        start_times = (np.arange(time_offset, num_frames*temporal_len,
                                 temporal_len)).tolist()

        self.list_outputs = list(zip([list_mode]*len(start_times), start_times,
                                 [temporal_len]*len(start_times)))

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["out"] = self.list_outputs

        return outputs


class PETListModeUnlistingInputSpec(BaseInterfaceInputSpec):

    list_inputs = traits.List(desc='List containing the time_offset and the '
                              'temporal frame length, as generated by '
                              'PrepareUnlistingInputs')


class PETListModeUnlistingOutputSpec(TraitedSpec):

    pet_sinogram = File(exists=True, desc='unlisted sinogram.')


class PETListModeUnlisting(BaseInterface):

    input_spec = PETListModeUnlistingInputSpec
    output_spec = PETListModeUnlistingOutputSpec

    def _run_interface(self, runtime):

        file_path = self.inputs.list_inputs[0]
        start = self.inputs.list_inputs[1]
        frame_len = self.inputs.list_inputs[2]
        end = start+frame_len
        print('Unlisting Frame {}'.format(str(start/frame_len)))
        cmd = (
            '{0} {1} 0 {2} 4 {3} {4} {5}'.format(
                list_mode_framing_path, file_path, str(frame_len), str(start),
                str(end), str(start/frame_len)))
        sp.check_output(cmd, shell=True)

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        start = self.inputs.list_inputs[0]
        frame_len = self.inputs.list_inputs[1]
        fname = self.inputs.list_mode

        _, _, ext = split_filename(fname)

        outputs["pet_sinogram"] = (
            os.getcwd()+'/Frame{0}{1}'.format(str(start/frame_len).zfill(5),
                                              ext))

        return outputs


class SSRBInputSpec(BaseInterfaceInputSpec):

    unlisted_sinogram = File(exists=True, desc='unlisted sinogram, output of '
                             'PETListModeUnlisting.')


class SSRBOutputSpec(TraitedSpec):

    ssrb_sinogram = File(exists=True, desc='Sinogram compressed using SSRB '
                         'algorithm. This will be the input of the PCA method '
                         'for motion detection')


class SSRB(BaseInterface):

    input_spec = SSRBInputSpec
    output_spec = SSRBOutputSpec

    def _run_interface(self, runtime):

        unlisted_sinogram = self.inputs.unlisted_sinogram
        basename = unlisted_sinogram.split('.')[0]
        new_ext = '.s'
        new_name = basename+new_ext
        os.rename(unlisted_sinogram, new_name)
        self.gen_interfiles(new_name)
        num_segs_to_combine = 1
        view_mash = 36
        do_ssrb_norm = 0
        cmd = ('SSRB {0}_ssrb {0}.hs {1} {2} {3}'
               .format(basename, num_segs_to_combine, view_mash, do_ssrb_norm))
        sp.check_output(cmd, shell=True)
        os.remove(basename+'.hs')

        return runtime

    def gen_interfiles(self, sinogram):

        file_pre = sinogram.split('/')[-1].split('.')[0]
        file_suf = '.hs'
        print('New file name: ' + file_pre + file_suf)
        with open(file_pre + file_suf, 'w') as new_file:
            with open(interfile_path) as old_file:
                for line in old_file:
                    new_file.write(
                        line.replace('name of data file := biograph_mmr.s\n',
                                     'name of data file := ' +
                                     sinogram.split('/')[-1]+'\n'))

        new_file.close()
        old_file.close()

    def _list_outputs(self):
        outputs = self._outputs().get()
        unlisted_sinogram = self.inputs.unlisted_sinogram
        basename = unlisted_sinogram.split('.')[0]

        outputs["ssrb_sinogram"] = basename+'_ssrb.s'

        return outputs


class MergeUnlistingOutputsInputSpec(BaseInterfaceInputSpec):

    sinograms = traits.List(desc='List of ssrb sinogram to merge into'
                            'one folder.')


class MergeUnlistingOutputsOutputSpec(TraitedSpec):

    sinogram_folder = Directory(desc='Directory containing all the compressed '
                                'sinograms.')


class MergeUnlistingOutputs(BaseInterface):

    input_spec = MergeUnlistingOutputsInputSpec
    output_spec = MergeUnlistingOutputsOutputSpec

    def _run_interface(self, runtime):

        sinograms = self.inputs.sinograms
        if os.path.isdir('PET_sinograms_for_PCA') is False:
            os.mkdir('PET_sinograms_for_PCA')
        for s in sinograms:
            shutil.move(s, 'PET_sinograms_for_PCA')

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["sinogram_folder"] = os.getcwd()+'/PET_sinograms_for_PCA'

        return outputs


class PreparePetDirInputSpec(BaseInterfaceInputSpec):

    pet_dir = Directory(exists=True, desc='Directory with the PET images to '
                        'use for motion correction.')
    image_orientation_check = traits.Bool(
        desc='This check is to assure that the PET images have the correct '
        'orientation (for example, using old version of e7tools the fineal PET'
        'images had wrong RL, AP and IS labels with respect to the actual '
        'image orientation). By default the pipeline checks the version of the'
        ' e7tools version used for the reconstrution in the PET header. If '
        'there is no information there and you are sure the orientation is '
        'correct then set this to True, otherwise recontruct the images with a'
        'new e7tools version. This software does not support old e7tools.',
        default=False)


class PreparePetDirOutputSpec(TraitedSpec):

    pet_dir_prepared = Directory(desc='Directory with PET images ready for the'
                                 ' motion correction.')


class PreparePetDir(BaseInterface):

    input_spec = PreparePetDirInputSpec
    output_spec = PreparePetDirOutputSpec

    def _run_interface(self, runtime):

        pet_dir = self.inputs.pet_dir
        image_orientation_check = self.inputs.image_orientation_check
        basename = 'frame'
        pet_images = sorted(
            glob.glob(pet_dir+'/{0}*.nii.gz'.format(basename)))

        if not pet_images:
            pet_images = sorted(
                glob.glob(pet_dir+'/{0}*.nii'.format(basename)))
        if pet_images:
            im = nib.load(pet_images[0])
            hd = im.header
            if 'New_e7tools' in hd['db_name']:
                image_orientation_check = True
                print ('New e7tool version detected.')
            pet_dicoms = sorted(glob.glob(pet_dir + '/Frame*'))
            if pet_dicoms and len(pet_images) != len(pet_dicoms):
                for f in pet_images:
                    os.remove(f)
                pet_images = []
        if not pet_images:
            pet_dicoms = sorted(glob.glob(pet_dir + '/Frame*'))
            if pet_dicoms:
                vol0 = sorted(glob.glob(pet_dicoms[0]+'/*'))[0]
                hd = pydicom.read_file(vol0)
                if ('e7tools' in hd.SoftwareVersions or
                        'syngo MR B20P' in hd.SoftwareVersions or
                        'syngo MR E11' in hd.SoftwareVersions):
                    image_orientation_check = True
                    print ('New e7tool version detected.')
                for dcm in pet_dicoms:
                    frame_num = dcm.split('/')[-1][5:]
                    cmd = ('mrconvert -force {0} {1}/{2}{3}.nii.gz'
                           .format(dcm, pet_dir, basename,
                                   str(frame_num).zfill(3)))
                    sp.check_output(cmd, shell=True)
                    cmd = ('fslreorient2std {0}/{1}{2}.nii.gz {0}/{1}{2}'
                           .format(pet_dir, basename, str(frame_num).zfill(3)))
                    sp.check_output(cmd, shell=True)
                    if frame_num == '0' and image_orientation_check:
                        im = nib.load('{0}/{1}{2}.nii.gz'.format(
                                pet_dir, basename, str(frame_num).zfill(3)))
                        hd = im.header
                        hd['db_name'] = 'New_e7tools'
                        nib.save(
                            im, '{0}/{1}{2}.nii.gz'.format(
                                pet_dir, basename, str(frame_num).zfill(3)))
                pet_images = sorted(glob.glob(
                    pet_dir+'/{0}*.nii.gz'.format(basename)))
            else:
                raise Exception("No PET images found in {0}!".format(pet_dir))
        if not image_orientation_check:
            raise Exception(
                "Could not find any e7tools version information in the PET "
                "header. If you are sure that the reconstructed PET images "
                "have correct orientation then specify image_orientation_check"
                "=True. Otherwise reconstruct your images with the new version"
                ". This software does not support the old e7tools version.")
        os.mkdir('pet_data')
        for f in pet_images:
            shutil.move(f, 'pet_data')

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["pet_dir_prepared"] = os.getcwd()+'/pet_data'

        return outputs


class PETFovCroppingInputSpec(BaseInterfaceInputSpec):

    pet_image = File(exists=True, desc='PET images to crop.')
#     ref_pet = File(exists=True, desc='Reference image to use to save the '
#                    'cropped PET. Usually is the output of fslroi command '
#                    'with the same cropping parameters.')
    x_min = traits.Int()
    x_size = traits.Int()
    y_min = traits.Int()
    y_size = traits.Int()
    z_min = traits.Int()
    z_size = traits.Int()


class PETFovCroppingOutputSpec(TraitedSpec):

    pet_cropped = File(exists=True, desc='Cropped PET')


class PETFovCropping(BaseInterface):

    input_spec = PETFovCroppingInputSpec
    output_spec = PETFovCroppingOutputSpec

    def _run_interface(self, runtime):

        pet_image = self.inputs.pet_image
#         ref = self.inputs.ref_pet
        x_min = self.inputs.x_min
        x_size = self.inputs.x_size
        y_min = self.inputs.y_min
        y_size = self.inputs.y_size
        z_min = self.inputs.z_min
        z_size = self.inputs.z_size
        _, basename, ext = split_filename(pet_image)
        outname = basename+'_crop'+ext
        pet = nib.load(pet_image)
        new_affine = np.copy(pet.affine)
        new_affine[:3, -1] = (pet.affine[:3, -1]-np.multiply(
            pet.header.get_zooms()[:3], (x_min, y_min, z_min)) *
            np.sign(pet.affine[:3, -1]))
        pet = pet.get_data()
        if len(pet.shape) == 3:
            pet_cropped = pet[x_min:x_min+x_size, y_min:y_min+y_size,
                              z_min:z_min+z_size]
        elif len(pet.shape) == 4:
            pet_cropped = pet[x_min:x_min+x_size, y_min:y_min+y_size,
                              z_min:z_min+z_size, :]
#         cmd = 'fslroi {} ref_roi 100 130 100 130 20 100'.format(im)
#         sp.check_output(cmd, shell=True)
#         ref = nib.load(ref)
        im2save = nib.Nifti1Image(pet_cropped, affine=new_affine)
        im2save.set_qform(new_affine, code='scanner')
        im2save.set_sform(new_affine, code='scanner')
        nib.save(im2save, outname)

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        pet_image = self.inputs.pet_image
        _, basename, ext = split_filename(pet_image)
        outname = basename+'_crop'+ext

        outputs["pet_cropped"] = os.getcwd()+'/'+outname

        return outputs


class CheckPetMCInputsInputSpec(BaseInterfaceInputSpec):

    pet_data = Directory(desc='Directory with the reconstructed PET images.')
    motion_mats = Directory(
        desc='Directory with the motion matrices calculated by the '
        'frame2reference pipeline')
    corr_factors = File(
        desc='Text file with the PET temporal correction factors used to '
        'generate the static PET motion corrected image.')
    reference = File(desc='Motion correction reference image.')


class CheckPetMCInputsOutputSpec(TraitedSpec):

    pet_images = traits.List(desc='List of PET images found in the PET recon'
                             ' directory.')
    motion_mats = traits.List(desc='List of motion matrices.')
    corr_factors = traits.List(desc='List of PET temporal correction factors.')
    pet2ref_mat = File(exists=True, desc='Matrix that transform images from '
                       'PET space to reference space.')


class CheckPetMCInputs(BaseInterface):

    input_spec = CheckPetMCInputsInputSpec
    output_spec = CheckPetMCInputsOutputSpec

    def _run_interface(self, runtime):

        dct = {}
        pet_data = sorted(glob.glob(self.inputs.pet_data+'/*.nii.gz'))
        motion_mats = sorted(glob.glob(
            self.inputs.motion_mats+'/*.txt'))
        reference = self.inputs.reference
        if isdefined(self.inputs.corr_factors):
            corr_factors = np.loadtxt(self.inputs.corr_factors).tolist()
        else:
            corr_factors = None
        if not pet_data:
            raise Exception('No images found in {}!'
                            .format(self.inputs.pet_cropped))
        elif (pet_data and (len(pet_data) != len(motion_mats))):
            raise Exception("The number of the PET images found in {0} is "
                            "different from that of the motion matrices found "
                            "in {1}. Please check."
                            .format(self.inputs.pet_data,
                                    self.inputs.motion_mats))
        else:
            pet_qform = self.get_qform(pet_data[0])
            ref_qform = self.get_qform(reference)
            ref_qform_inv = np.linalg.inv(ref_qform)
            pet2ref = np.dot(ref_qform_inv, pet_qform)
            np.savetxt('pet2ref.mat', pet2ref)
            dct['pet_data'] = pet_data
            dct['motion_mats'] = motion_mats
            if (corr_factors is not None and
                    (len(pet_data) == len(corr_factors))):
                dct['corr_factors'] = corr_factors
            elif corr_factors is None:
                dct['corr_factors'] = []
            if (corr_factors is not None and
                    (len(pet_data) != len(corr_factors))):
                raise Exception(
                    "The number of the PET images found in {0} is {1} and it "
                    "is different from that of the PET correction factors"
                    "which is {2}. Please check."
                    .format(self.inputs.pet_data, len(pet_data),
                            len(corr_factors)))
            self.dct = dct

        return runtime

    def get_qform(self, image):

        cmd = 'mrinfo {}'.format(image)
        hd = (sp.check_output(cmd, shell=True)).decode('utf-8')
        i = [n for n, el in enumerate(hd.split('\n')) if 'Transform' in el][0]
        mat = []
        for j in range(3):
            if j+i == i:
                mat.append([float(x) for x in hd.split('\n')[j+i].split()[1:]])
            else:
                mat.append([float(x) for x in hd.split('\n')[j+i].split()])
        mat.append([0, 0, 0, 1])
        return np.asarray(mat)

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["pet_images"] = self.dct['pet_data']
        outputs["motion_mats"] = self.dct['motion_mats']
        outputs["corr_factors"] = self.dct['corr_factors']
        outputs["pet2ref_mat"] = os.getcwd()+'/pet2ref.mat'

        return outputs


class PetImageMotionCorrectionInputSpec(BaseInterfaceInputSpec):

    pet_image = File(desc='Directory with the fov cropped PET images.')
    motion_mat = File(desc='Directory with the outputs from the MR-based '
                      'motion detection pipeline.')
    structural_image = File(desc='If provided, the final PET mc image will be '
                            'aligned to this image.', default=None)
    corr_factor = traits.Float()
    pet2ref_mat = File(exists=True)
    structural2ref_regmat = File(default=None)


class PetImageMotionCorrectionOutputSpec(TraitedSpec):

    pet_mc_image = File(desc='Motin corrected static PET results.')
    pet_no_mc_image = File(desc='Motin corrected static PET results.')


class PetImageMotionCorrection(BaseInterface):

    input_spec = PetImageMotionCorrectionInputSpec
    output_spec = PetImageMotionCorrectionOutputSpec

    def _run_interface(self, runtime):

        motion_mat = np.loadtxt(self.inputs.motion_mat)
        structural_image = self.inputs.structural_image
        if isdefined(self.inputs.structural2ref_regmat):
            structural2ref_regmat = np.loadtxt(
                self.inputs.structural2ref_regmat)
        pet2ref_mat = np.loadtxt(self.inputs.pet2ref_mat)
        pet_image = self.inputs.pet_image
        if isdefined(self.inputs.corr_factor):
            corr_factor = self.inputs.corr_factor
        else:
            corr_factor = 1

        ref2pet_mat = np.linalg.inv(pet2ref_mat)
        if structural_image:
            ref2pet_mat = np.linalg.inv(structural2ref_regmat)
            out_basename = 'al2Struct'
        else:
            out_basename = 'al2Ref'
        basename = pet_image.split('/')[-1].split('.')[0]
        outname = '{0}_{1}'.format(basename, out_basename)
        motion_mat_inv = np.linalg.inv(motion_mat)
        transformation_mat = np.dot(ref2pet_mat,
                                    np.dot(motion_mat_inv, pet2ref_mat))
        np.savetxt('transformation.mat', transformation_mat)

        if structural_image:
            self.applyxfm(pet_image, structural_image, 'transformation.mat',
                          outname+'_mc')
        else:
            self.applyxfm(pet_image, pet_image, 'transformation.mat',
                          outname+'_mc')
        self.apply_temporal_correction(outname+'_mc', corr_factor,
                                       outname+'_mc_corr')
        self.apply_temporal_correction(pet_image, corr_factor,
                                       outname+'_no_mc_corr')
        self.out_basename = out_basename

        return runtime

    def apply_temporal_correction(self, image, corr_factor, out_name):

        cmd = ('fslmaths {0}.nii.gz -mul {1} {2}'
               .format(image, corr_factor, out_name))
        sp.check_output(cmd, shell=True)

    def applyxfm(self, in_file, ref, mat, outname):

        applyxfm = fsl.FLIRT()
        applyxfm.inputs.in_file = in_file
        applyxfm.inputs.reference = ref
        applyxfm.inputs.apply_xfm = True
        applyxfm.inputs.in_matrix_file = mat
        applyxfm.inputs.out_file = outname + '.nii.gz'
        applyxfm.run()

    def extract_qform(self, image):

        cmd = 'fslhd {}'.format(image)
        image_info = (sp.check_output(cmd, shell=True)).decode('utf-8')
        image_info = image_info.strip().split('\n')
        qform = np.eye(4)
        for i, line in enumerate(image_info):
            if 'qto_xyz:1' in line:
                qform[0, -1] = abs(float(
                    image_info[i].split(':1')[-1].strip().split()[-1]))
                qform[1, -1] = abs(float(
                    image_info[i+1].split(':2')[-1].strip().split()[-1]))
                qform[2, -1] = abs(float(
                    image_info[i+2].split(':3')[-1].strip().split()[-1]))
                break

        return qform

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["pet_mc_image"] = glob.glob(
            os.getcwd()+'/*{}_mc_corr.nii.gz'
            .format(self.out_basename))[0]
        outputs["pet_no_mc_image"] = glob.glob(
            os.getcwd()+'/*no_mc_corr.nii.gz')[0]
        return outputs


class StaticPETImageGenerationInputSpec(BaseInterfaceInputSpec):

    pet_mc_images = traits.List()
    pet_no_mc_images = traits.List()


class StaticPETImageGenerationOutputSpec(TraitedSpec):

    static_mc = File()
    static_no_mc = File()


class StaticPETImageGeneration(BaseInterface):

    input_spec = StaticPETImageGenerationInputSpec
    output_spec = StaticPETImageGenerationOutputSpec

    def _run_interface(self, runtime):

        pet_mc_images = self.inputs.pet_mc_images
        pet_no_mc_images = self.inputs.pet_no_mc_images

        self.frames_sum('mc_corr', pet_mc_images)
        self.frames_sum('no_mc_corr', pet_no_mc_images)

        return runtime

    def frames_sum(self, outname, images):

        cmd = 'fslmaths '
        for i, frame in enumerate(images):
            if i != len(images)-1:
                cmd = cmd + '{0} -add '.format(frame)
            else:
                cmd = (cmd+'{0} static_PET_{1}'.format(frame, outname))
        sp.check_output(cmd, shell=True)

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["static_mc"] = os.getcwd()+'/static_PET_mc_corr.nii.gz'
        outputs["static_no_mc"] = os.getcwd()+'/static_PET_no_mc_corr.nii.gz'
        return outputs
